"""Mesh a polygon."""


from pathlib import PurePath
from typing import Optional

import gmsh

from .circular_iterable import circular_pairwise
from .geometry import Geometry
from .gmsh_context_manager import GmshContextManager, GmshOptions
from .mesh_border import _mesh_border
from .polygon import Polygon


def mesh(
    geom: Geometry, mesh_size: float, gmsh_options: Optional[GmshOptions] = None
) -> Optional[PurePath]:
    """Mesh a geometry.

    Parameters
    ----------
    geom : Geometry
    mesh_size : float
    gmsh_options: GmshOptions, optional

    """

    if gmsh_options is None:
        gmsh_options = GmshOptions()

    with GmshContextManager(gmsh_options):
        loop_polygons = [_mesh_polygon(polygon, mesh_size) for polygon in geom.polygons]
        loop_border = _mesh_border(geom.border, mesh_size)

        s_vac = gmsh.model.geo.addPlaneSurface([loop_border, *loop_polygons])
        gmsh.model.addPhysicalGroup(dim=2, tags=[s_vac], name="Vacuum")

    return gmsh_options.filename


def _mesh_polygon(polygon: Polygon, h: float) -> int:
    """Mesh polygon."""
    points = [
        gmsh.model.geo.addPoint(corner.c[0], corner.c[1], 0, h)
        for corner in polygon.corners
    ]

    lines = [gmsh.model.geo.addLine(A, B) for A, B in circular_pairwise(points)]

    loop = gmsh.model.geo.addCurveLoop(lines)

    surface = gmsh.model.geo.addPlaneSurface([loop])
    gmsh.model.addPhysicalGroup(dim=2, tags=[surface], name=polygon.name)

    return loop



"""GMSH as a context manager."""


from contextlib import AbstractContextManager
from dataclasses import dataclass, field
from pathlib import PurePath
from typing import Any, NamedTuple, Optional, TypeAlias, Union

import gmsh

OptionalPathLike: TypeAlias = Union[None, str, PurePath]


@dataclass(slots=True)
class GmshOptions:
    """GMSH Options.

    http://gmsh.info/doc/texinfo/gmsh.html#Gmsh-options

    Attributes
    ----------
    element_order : int, default=1
    terminal : bool, default=False
    gui : bool,default=False
    filename : OptionalPathLike, optional
    hide_model_entities : bool, default=True
    additional_options : dict[str, Any], optional

    """

    gui: bool = False
    filename: Optional[PurePath] = None
    hide_model_entities: bool = True
    option_value: dict[str, Any] = field(default_factory=dict)

    def __init__(
        self,
        *,
        element_order: int = 1,
        terminal: bool = False,
        gui: bool = False,
        filename: OptionalPathLike = None,
        hide_model_entities: bool = True,
        additional_options: Optional[dict[str, Any]] = None,
    ) -> None:
        if element_order < 1:
            raise ValueError("Element order must be an integer greater or equal to 1.")

        self.gui = gui
        self.filename = PurePath(filename) if filename is not None else None
        self.hide_model_entities = hide_model_entities

        option_value = {
            "General.Terminal": int(terminal),
            "General.SmallAxes": 0,
            "Geometry.CopyMeshingMethod": 1,
            "Mesh.ElementOrder": element_order,
            "Mesh.TransfiniteTri": 1,
            "Mesh.Smoothing": 0,
            "Mesh.SurfaceFaces": 1,
            "Mesh.ColorCarousel": 2,
        }
        if additional_options is not None:
            option_value = {**option_value, **additional_options}

        self.option_value = option_value

    def __str__(self) -> str:
        data = [
            ("Gmsh options", ""),
            ("  Launch GMSH GUI", self.gui),
            ("  filename", self.filename),
            ("  Hide model entities", self.hide_model_entities),
            ("#", ""),
            ("Gmsh options", ""),
        ]
        data.extend(((f"  {key}", val) for key, val in self.option_value.items()))
        return _pretty_print(data)


def _pretty_print(data: list[tuple[str, Any]]) -> str:
    """Pretty print."""
    max_len = max(len(d[0]) for d in data) + 1
    return "".join(
        f"{d[0]} {'.' * (max_len - len(d[0]))} {d[1]}\n" if d[0][0] != "#" else "\n"
        for d in data
    )


@dataclass(frozen=True, slots=True)
class GmshContextManager(AbstractContextManager):
    """Context manager for GMSH."""

    # pylint: disable=too-many-instance-attributes

    gmsh_options: GmshOptions

    def __enter__(self):
        # Initialize the Gmsh API.
        gmsh.initialize()

        for key, val in self.gmsh_options.option_value.items():
            gmsh.option.setNumber(key, val)

        # Add a new model and set it as the current model.
        if self.gmsh_options.filename is None:
            gmsh.model.add("mesh")
        else:
            gmsh.model.add(self.gmsh_options.filename.stem)

    def __exit__(self, exc_type, exc_value, exc_tb):
        # Synchronize the built-in CAD representation with the current Gmsh model.
        gmsh.model.geo.synchronize()

        # Generate a mesh of the current model, up to dimension dim 2.
        gmsh.model.mesh.generate(2)

        _set_options_graphical(self.gmsh_options.hide_model_entities)

        if self.gmsh_options.gui:
            # Create and run the FLTK graphical user interface.
            gmsh.fltk.run()

        if self.gmsh_options.filename is not None:
            # Write a file. The export format is determined by the file extension.
            match self.gmsh_options.filename.suffix:
                case ".msh":
                    gmsh.write(str(self.gmsh_options.filename))
                case _:
                    gmsh.fltk.initialize()
                    gmsh.write(str(self.gmsh_options.filename))
                    gmsh.fltk.finalize()

        # Finalize the Gmsh API.
        gmsh.finalize()

        if exc_type is not None:
            print(f"\n{exc_type}")

        if exc_value is not None:
            print(f"\n{exc_value}")

        if exc_tb is not None:
            print(f"\n{exc_tb}")


INT_TO_STR = {
    0: "Zero",
    1: "One",
    2: "Two",
    3: "Three",
    4: "Four",
    5: "Five",
    6: "Six",
    7: "Seven",
    8: "Eight",
    9: "Nine",
    10: "Ten",
    11: "Eleven",
    12: "Twelve",
    13: "Thirteen",
    14: "Fourteen",
    15: "Fifteen",
    16: "Sixteen",
    17: "Seventeen",
    18: "Eighteen",
    19: "Nineteen",
}


class Color(NamedTuple):
    """Color representation."""

    # matplotlib color name
    mpl: str

    # RGBA representation
    r: int
    g: int
    b: int
    a: int = 255


C_POLYGON = Color("C2", 44, 160, 44)
C_VACUUM = Color("C0", 31, 119, 180)
C_PML = Color("C1", 255, 127, 14)


def _set_options_graphical(hide_model_Entities: bool) -> None:
    """Set."""

    # Hide the model entities of dimensions 0 and 1.
    if hide_model_Entities:
        gmsh.model.setVisibility(gmsh.model.getEntities(0), 0)
        gmsh.model.setVisibility(gmsh.model.getEntities(1), 0)

    name_to_color = {
        gmsh.model.getPhysicalName(dim=d, tag=t): f"Mesh.Color.{INT_TO_STR[t]}"
        for (d, t) in gmsh.model.getPhysicalGroups(dim=2)
    }

    for name, color in name_to_color.items():
        if name == "Vacuum":
            gmsh.option.setColor(color, C_VACUUM.r, C_VACUUM.g, C_VACUUM.b, C_VACUUM.a)
        elif name == "PML":
            gmsh.option.setColor(color, C_PML.r, C_PML.g, C_PML.b, C_PML.a)
        else:
            gmsh.option.setColor(
                color, C_POLYGON.r, C_POLYGON.g, C_POLYGON.b, C_POLYGON.a
            )



"""Config file for sphinx."""

try:
    import lostinmsh
except ImportError:
    import os
    import sys

    sys.path.insert(0, os.path.abspath("../"))

    import lostinmsh

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "lostinmsh"
copyright = f"2023, {lostinmsh.__author__}"
author = lostinmsh.__author__
release = lostinmsh.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

# -- Extensions --------------------------------------------------------------
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx.ext.coverage",
    "sphinx.ext.githubpages",
    "sphinx_rtd_theme",
    "sphinx_copybutton",
    "nbsphinx",
]

nbsphinx_execute = "never"

# "sphinx.ext.linkcode",
# def linkcode_resolve(domain, info):
#     if domain != "py":
#         return None
#     if not info["module"]:
#         return None
#     filename = info["module"].replace(".", "/")
#     return f"https://github.com/zmoitier/lostinmsh/{filename}.py"



"""T-conform mesh a polygon."""


from dataclasses import dataclass
from pathlib import PurePath
from typing import Optional, TypeAlias

import gmsh
from numpy import cos, linspace, pi, sin, sqrt

from .circular_iterable import circular_pairwise
from .geometry import Geometry
from .gmsh_context_manager import GmshContextManager, GmshOptions
from .mesh_border import _mesh_border
from .polygon import Angle, Corner, Polygon


@dataclass(frozen=True, kw_only=True, slots=True)
class CornerTag:
    """Class for corner's tag."""

    pts_cav: list[int]
    pts_vac: list[int]
    lines_cav: list[int]
    lines_vac: list[int]


@dataclass(frozen=True, kw_only=True, slots=True)
class EgdeTag:
    """Class for edge's tag."""

    line_cav: int
    line_vac: int


SurfaceTags: TypeAlias = list[int]


def mesh_loc_struct(
    geometry: Geometry,
    mesh_size: float,
    gmsh_options: Optional[GmshOptions],
    *,
    corner_radius_shrink: float = 0.75,
    corner_geometric_coef: float = 1.5,
) -> Optional[PurePath]:
    """T-confrom mesh a polygon.

    Parameters
    ----------
    geom : Geometry
    mesh_size : float
    gmsh_options: GmshOptions, optional
    corner_radius_shrink : float, default=0.75
    corner_geometric_coef : float, default=1.5

    """

    if gmsh_options is None:
        gmsh_options = GmshOptions()

    corner_radius = geometry.max_corner_radius() * corner_radius_shrink

    with GmshContextManager(gmsh_options):
        loop_polygons = []
        surfaces_vac = []
        for polygon in geometry.polygons:
            l_poly, s_poly, s_vac = _tmesh_polygon(
                polygon, corner_radius, mesh_size, corner_geometric_coef
            )
            loop_polygons.append(l_poly)
            surfaces_vac.extend(s_vac)

            gmsh.model.addPhysicalGroup(dim=2, tags=s_poly, name=polygon.name)

        loop_border = _mesh_border(geometry.border, mesh_size)

        surfaces_vac.append(
            gmsh.model.geo.addPlaneSurface([loop_border, *loop_polygons])
        )

        gmsh.model.addPhysicalGroup(dim=2, tags=surfaces_vac, name="Vacuum")

    return gmsh_options.filename


def _tmesh_polygon(
    polygon: Polygon, corner_radius: float, h: float, corner_coef: float
) -> tuple[int, SurfaceTags, SurfaceTags]:
    """T-coform mesh of a polygon."""

    ϕ = polygon.get_elementary_angle()
    r = min(corner_radius, h / (2 * sin(ϕ.value / 2)))
    h_corner = sqrt(h * r * 2 * sin(ϕ.value / 2))

    corners_tag = []
    surfaces_cav = []
    surfaces_vac = []
    N_corner = max(2, round(1 + r / h))

    for corner in polygon.corners:
        c_tag, s_cav, s_vac = _mesh_corner(corner, r, ϕ, h, N_corner, corner_coef)
        corners_tag.append(c_tag)
        surfaces_cav.extend(s_cav)
        surfaces_vac.extend(s_vac)

    edges_tag = []
    for l, (C, D) in zip(polygon.lengths, circular_pairwise(corners_tag)):
        N_edge = max(2, round(1 + (l - 2 * r) / h_corner))
        e_tag, s_cav, s_vac = _mesh_edge(C, D, N_edge)
        edges_tag.append(e_tag)
        surfaces_cav.extend(s_cav)
        surfaces_vac.extend(s_vac)

    loop_inn, loop_out = _get_loops(corners_tag, edges_tag)

    surfaces_cav.append(gmsh.model.geo.addPlaneSurface([loop_inn]))

    return (loop_out, surfaces_cav, surfaces_vac)


def _mesh_corner(
    corner: Corner, r: float, ϕ: Angle, h: float, N: int, f: float
) -> tuple[CornerTag, list[int], list[int]]:
    """Mesh corner."""

    p, q = _compute_pq(corner.angle, ϕ)

    C = gmsh.model.geo.addPoint(corner.c[0], corner.c[1], 0, h)
    points = [
        gmsh.model.geo.addPoint(*to_global(corner, r, θ), 0, h)
        for θ in linspace(0, 2 * pi, num=p + q, endpoint=False)
    ]

    lines_rad = [gmsh.model.geo.addLine(C, A) for A in points]
    lines_ang = [gmsh.model.geo.addLine(A, B) for A, B in circular_pairwise(points)]

    for line in lines_rad:
        gmsh.model.geo.mesh.setTransfiniteCurve(line, N, coef=f)
    for line in lines_ang:
        gmsh.model.geo.mesh.setTransfiniteCurve(line, N)

    loops = []
    for i, j in circular_pairwise(range(p + q)):
        loops.append(
            gmsh.model.geo.addCurveLoop([lines_rad[i], lines_ang[i], -lines_rad[j]])
        )

    surfaces = []
    for loop in loops:
        s = gmsh.model.geo.addPlaneSurface([loop])
        gmsh.model.geo.mesh.setTransfiniteSurface(s)
        surfaces.append(s)

    return (
        CornerTag(
            pts_cav=points[:p],
            pts_vac=points[p:],
            lines_cav=lines_ang[:p],
            lines_vac=lines_ang[p:],
        ),
        surfaces[:p],
        surfaces[p:],
    )


def _compute_pq(angle: Angle, ϕ: Angle) -> tuple[int, int]:
    "Compute p and q such that a / b = 2p / (p+q) and p, q >= 2."
    return (int(angle / ϕ), int((2 - angle) / ϕ))


def to_global(corner: Corner, r: float, θ: float) -> tuple[float, float]:
    """Transform the local corner coordinate to global coordinate."""
    cosθ, sinθ = cos(θ), sin(θ)
    return (
        corner.c[0] + r * (cosθ * corner.v1[0] - sinθ * corner.v1[1]),
        corner.c[1] + r * (sinθ * corner.v1[0] + cosθ * corner.v1[1]),
    )


def _mesh_edge(
    C: CornerTag, D: CornerTag, N: int
) -> tuple[EgdeTag, SurfaceTags, SurfaceTags]:
    """Mesh edge."""
    interface = gmsh.model.geo.addLine(C.pts_cav[0], D.pts_vac[0])
    line_cav = gmsh.model.geo.addLine(D.pts_cav[-1], C.pts_cav[1])
    line_vac = gmsh.model.geo.addLine(C.pts_vac[-1], D.pts_vac[1])

    for line in (interface, line_cav, line_vac):
        gmsh.model.geo.mesh.setTransfiniteCurve(line, N)

    surface_cav = gmsh.model.geo.addPlaneSurface(
        [
            gmsh.model.geo.addCurveLoop(
                [interface, -D.lines_cav[-1], line_cav, -C.lines_cav[0]]
            )
        ]
    )
    gmsh.model.geo.mesh.setTransfiniteSurface(surface_cav)

    surface_vac = gmsh.model.geo.addPlaneSurface(
        [
            gmsh.model.geo.addCurveLoop(
                [-interface, -D.lines_vac[0], line_vac, -C.lines_vac[-1]]
            )
        ]
    )
    gmsh.model.geo.mesh.setTransfiniteSurface(surface_vac)

    return (EgdeTag(line_cav=line_cav, line_vac=line_vac), [surface_cav], [surface_vac])


def _get_loops(
    corners_tag: list[CornerTag], edges_tag: list[EgdeTag]
) -> tuple[int, int]:
    """Get loop inner and outer."""
    loop_inn = []
    loop_out = []
    for c, e in zip(corners_tag, edges_tag):
        loop_inn.extend(c.lines_cav[1:-1])
        loop_inn.append(e.line_cav)

        loop_out.extend(c.lines_vac[1:-1])
        loop_out.append(e.line_vac)

    return (
        gmsh.model.geo.addCurveLoop([-l for l in loop_inn]),
        gmsh.model.geo.addCurveLoop(loop_out),
    )



"""GMSH toolbox for locally structured meshes on polygon."""

__version__ = "0.0.2"

__author__ = "Zoïs Moitier, Camille Carvalho"

__all__ = [
    "Polygon",
    "Geometry",
    "Circular",
    "Rectangular",
    "AutoCircular",
    "AutoRectangular",
    "GmshOptions",
    "mesh",
    "mesh_loc_struct",
    "plot_geometry",
    "plot_polygon",
]

from .border import AutoCircular, AutoRectangular, Circular, Rectangular
from .geometry import Geometry
from .gmsh_context_manager import GmshOptions
from .mesh import mesh
from .mesh_lost import mesh_loc_struct
from .plot import plot_geometry, plot_polygon
from .polygon import Polygon



"""Helper functions for circular iteration."""


from collections.abc import Collection
from itertools import cycle, islice
from typing import Iterable


def circular_pairwise(collection: Collection, start: int = 0) -> Iterable:
    """Circular pairwise."""
    end = start + len(collection)
    return zip(
        islice(cycle(collection), start, end),
        islice(cycle(collection), start + 1, None),
    )


def circular_triplewise(collection: Collection, start: int = 0) -> Iterable:
    """Circular triplewise."""
    end = start + len(collection)
    return zip(
        islice(cycle(collection), start, end),
        islice(cycle(collection), start + 1, None),
        islice(cycle(collection), start + 2, None),
    )



"""Helper function for meshes function."""


import gmsh
from numpy.typing import NDArray

from .border import Border, Circular, Rectangular
from .circular_iterable import circular_pairwise


def _mesh_border(border: Border, mesh_size: float) -> int:
    """Mesh the exterior."""

    if isinstance(border, Circular):
        return _mesh_circular(border, mesh_size)

    if isinstance(border, Rectangular):
        return _mesh_rectangular(border, mesh_size)

    raise ValueError("Unknown border shape.")


def _mesh_circular(circ: Circular, mesh_size: float) -> int:
    """Mesh circular border."""
    loop_inn = _loop_circle(circ.center, circ.radius, mesh_size, "Sigma")

    if circ.thickness is not None:
        loop_out = _loop_circle(
            circ.center, circ.radius + circ.thickness, mesh_size, "Sigma"
        )
        s_pml = gmsh.model.geo.addPlaneSurface([loop_out, loop_inn])
        gmsh.model.addPhysicalGroup(dim=2, tags=[s_pml], name="PML")

    return loop_inn


def _mesh_rectangular(rect: Rectangular, mesh_size: float) -> int:
    """Mesh rectangular border."""
    loop_inn = _loop_rectangle(
        rect.center, rect.half_width, rect.half_height, mesh_size, "Sigma"
    )

    if rect.thickness is not None:
        loop_out = _loop_rectangle(
            rect.center,
            rect.half_width + rect.thickness,
            rect.half_height + rect.thickness,
            mesh_size,
            "Sigma",
        )
        s_pml = gmsh.model.geo.addPlaneSurface([loop_out, loop_inn])
        gmsh.model.addPhysicalGroup(dim=2, tags=[s_pml], name="PML")

    return loop_inn


def _loop_circle(center: NDArray, radius: float, h: float, name: str) -> int:
    """Return loop corresponding to a circle."""
    cx, cy = center

    C = gmsh.model.geo.addPoint(cx, cy, 0, h)
    pts = [
        gmsh.model.geo.addPoint(cx + x, cy + y, 0, h)
        for x, y in zip([radius, 0, -radius, 0], [0, radius, 0, -radius])
    ]

    lines = [gmsh.model.geo.addCircleArc(A, C, B) for A, B in circular_pairwise(pts)]
    gmsh.model.addPhysicalGroup(dim=1, tags=lines, name=name)

    return gmsh.model.geo.addCurveLoop(lines)


def _loop_rectangle(
    center: NDArray, half_width: float, half_height: float, h: float, name: str
) -> int:
    """Return loop corresponding to a square."""
    cx, cy = center

    pts = [
        gmsh.model.geo.addPoint(cx + x, cy + y, 0, h)
        for x, y in zip(
            [half_width, -half_width, -half_width, half_width],
            [half_height, half_height, -half_height, -half_height],
        )
    ]

    lines = [gmsh.model.geo.addLine(A, B) for A, B in circular_pairwise(pts)]
    gmsh.model.addPhysicalGroup(dim=1, tags=lines, name=name)

    return gmsh.model.geo.addCurveLoop(lines)



"""Find the smallest distance of a set of points.

https://www.geeksforgeeks.org/closest-pair-of-points-using-divide-and-
conquer- algorithm/ https://stackoverflow.com/questions/39681407/find-
minimum-distance-between- points
https://sites.cs.ucsb.edu/~suri/cs235/ClosestPair.pdf

"""


from numpy import absolute, argsort, inf, less, where
from numpy.linalg import norm
from numpy.typing import NDArray


def min_dist_naive(points: NDArray) -> float:
    """Smallest distance of a set of points."""
    n = points.shape[0]

    min_d = inf
    for i in range(n):
        P = points[i, :]
        for j in range(i + 1, n):
            Q = points[j, :]
            if (d := norm(P - Q)) < min_d:
                min_d = float(d)
    return min_d


def min_dist(points: NDArray) -> float:
    """Smallest distance of a set of points."""
    idx_sort_x = argsort(points[:, 0])
    return _min_dist_rec(points[idx_sort_x, :])


def _min_dist_rec(points: NDArray) -> float:
    """Smallest distance square of a set of points."""
    n = points.shape[0]

    if n <= 3:
        return min_dist_naive(points)

    idx_mid = n // 2
    M = points[idx_mid, :]

    dl = _min_dist_rec(points[:idx_mid])
    dr = _min_dist_rec(points[idx_mid:])
    d = min(dl, dr)

    idx_strip = where(less(absolute(points[:, 0] - M[0]), d))[0]

    return min(d, _min_dist_strip(points[idx_strip, :], d))


def _min_dist_strip(points: NDArray, d: float) -> float:
    """Smallest distance strip."""
    n = points.shape[0]
    min_d = d

    idx_sort_y = argsort(points[:, 1])

    for i in range(n):
        P = points[idx_sort_y[i], :]
        for j in range(i + 1, n):
            Q = points[idx_sort_y[j], :]
            if Q[1] - P[1] >= min_d:
                break
            if (t := norm(P - Q)) < min_d:
                min_d = float(t)

    return min_d



"""Class for boundary."""


from dataclasses import dataclass
from typing import Optional

from numpy.linalg import norm
from numpy.typing import NDArray

from .smallest_boundary import smallest_circle, smallest_rectangle


@dataclass(kw_only=True, slots=True)
class Border:
    """Border class.

    Attributes
    ----------
    center : NDArray
    thickness : float, optional

    """

    center: NDArray
    thickness: Optional[float] = None

    def dist_to_inner_boundary(self, points: NDArray) -> float:
        """Sign distance to the inner boundary.

        It is positive is all the points are inside and it is negative
        if at least one point is outside.

        Parameters
        ----------
        points : NDArray
            list of points should be an array of shape (N, 2) with N ≥ 1.

        """
        raise NotImplementedError()


@dataclass(kw_only=True, slots=True)
class Circular(Border):
    """Circular boundary."""

    radius: float

    def dist_to_inner_boundary(self, points: NDArray) -> float:
        return self.radius - norm(points - self.center, axis=1).max()


@dataclass(kw_only=True, slots=True)
class Rectangular(Border):
    """Rectangular boundary."""

    half_width: float
    half_height: float

    def dist_to_inner_boundary(self, points: NDArray) -> float:
        pts = points - self.center
        return min(
            self.half_width - pts[:, 0].max(), self.half_height - pts[:, 1].max()
        )


@dataclass(kw_only=True, slots=True)
class AutoBorder:
    """Auto border.

    Attributes
    ----------
    border_factor : float
    thickness_factor : float, optional

    """

    border_factor: float
    thickness_factor: Optional[float] = None

    def get_border(self, points: NDArray) -> Border:
        """Auto set.

        Parameters
        ----------
        points : NDArray
            list of points should be an array of shape (N, 2) with N ≥ 1.

        """
        raise NotImplementedError()


@dataclass(kw_only=True, slots=True)
class AutoCircular(AutoBorder):
    """Auto circular.

    Attributes
    ----------
    border_factor : float
    thickness_factor :float, optional

    """

    def get_border(self, points: NDArray) -> Circular:
        center, radius = smallest_circle(points)

        r0 = radius * (1 + self.border_factor)
        if self.thickness_factor is not None:
            thickness = radius * self.thickness_factor
        else:
            thickness = None

        return Circular(center=center, radius=r0, thickness=thickness)


@dataclass(kw_only=True, slots=True)
class AutoRectangular(AutoBorder):
    """Auto rectangular.

    Attributes
    ----------
    border_factor : float
    thickness_factor : float, optional

    """

    def get_border(self, points: NDArray) -> Rectangular:
        center, lengths = smallest_rectangle(points)

        r = float(norm(lengths))
        l0 = lengths + r * self.border_factor
        if self.thickness_factor is not None:
            thickness = r * self.thickness_factor
        else:
            thickness = None

        return Rectangular(
            center=center, half_width=l0[0], half_height=l0[1], thickness=thickness
        )



"""Plot polygon and geometry."""


from numpy import pi

from .border import Circular, Rectangular
from .geometry import Geometry
from .gmsh_context_manager import C_PML, C_POLYGON, C_VACUUM
from .mesh_lost import to_global
from .polygon import Polygon
from .smallest_boundary import smallest_circle

ENABLE_PLOT = True
try:
    import matplotlib.pyplot as plt
    from matplotlib.patches import Circle as mpl_Circle
    from matplotlib.patches import Polygon as mpl_Polygon
    from matplotlib.patches import Rectangle as mpl_Rectangle
except ImportError:
    ENABLE_PLOT = False


def plot_polygon(polygon: Polygon, ax=None) -> None:
    """Plot polygon.

    Parameters
    ----------
    polygon : Polygon
    ax : plt.Axes, optional

    """

    if ENABLE_PLOT:
        _plot_polygon(polygon, ax=ax)
    else:
        raise ModuleNotFoundError(
            "You need to install matplotlib to use this function."
        )


def _plot_polygon(polygon: Polygon, ax=None) -> None:
    """Plot polygon."""

    if ax is None:
        _, ax = plt.subplots()

    ax.add_patch(
        mpl_Polygon(
            polygon.get_vertices(),
            linewidth=2,
            fill=False,
            color=C_POLYGON.mpl,
            zorder=2,
            label=polygon.name,
        )
    )

    l_min = polygon.lengths.min()
    for corner in polygon.corners:
        if corner.angle < 1:
            angle = pi + corner.angle.value / 2
        else:
            angle = corner.angle.value / 2
        pts = to_global(corner, 0.15 * l_min, angle)
        ax.text(pts[0], pts[1], corner.angle.latex(), va="center", ha="center")

    center, radius = smallest_circle(polygon.get_vertices())
    ax.add_patch(mpl_Circle(center, 1.2 * radius, fill=False, linewidth=0))

    ax.axis("equal")
    ax.grid(zorder=1)
    ax.legend(loc=1)


def plot_geometry(geometry: Geometry, ax=None) -> None:
    """Plot geometry.

    Parameters
    ----------
    geometry : Geometry
    ax : plt.Axes, optional

    """

    if ENABLE_PLOT:
        _plot_geometry(geometry, ax=ax)
    else:
        raise ModuleNotFoundError(
            "You need to install matplotlib to use this function."
        )


def _plot_geometry(geometry: Geometry, ax=None) -> None:
    """Plot geometry."""

    if ax is None:
        _, ax = plt.subplots()

    for polygon in geometry.polygons:
        ax.add_patch(
            mpl_Polygon(
                polygon.get_vertices(),
                linewidth=2,
                color=C_POLYGON.mpl,
                zorder=4,
                label=polygon.name,
            )
        )

    ax.plot([geometry.border.center[0]], [geometry.border.center[1]], "ko")

    options_vac = {
        "linewidth": 2,
        "linestyle": "--",
        "edgecolor": "k",
        "facecolor": C_VACUUM.mpl,
        "zorder": 3,
        "label": "Vaccum",
    }
    options_pml = {
        "linewidth": 2,
        "linestyle": "-.",
        "edgecolor": "k",
        "facecolor": C_PML.mpl,
        "zorder": 2,
        "label": "PML",
    }

    if isinstance(geometry.border, Circular):
        _add_circular(ax, geometry.border, options_vac, options_pml)
    elif isinstance(geometry.border, Rectangular):
        _add_rectangular(ax, geometry.border, options_vac, options_pml)
    else:
        raise ValueError("Unknown border shape.")

    ax.axis("equal")
    ax.grid(zorder=1)
    ax.legend(loc=1)


def _add_circular(ax, circ: Circular, options_vac, options_pml):
    """Add circular border."""
    ax.add_patch(mpl_Circle(circ.center, circ.radius, **options_vac))

    if circ.thickness is not None:
        ax.add_patch(
            mpl_Circle(circ.center, circ.radius * (1 + circ.thickness), **options_pml)
        )


def _add_rectangular(ax, rect: Rectangular, options_vac, options_pml):
    """Add circular border."""
    x, y = rect.center
    ax.add_patch(
        mpl_Rectangle(
            (x - rect.half_width, y - rect.half_height),
            2 * rect.half_width,
            2 * rect.half_height,
            **options_vac,
        )
    )

    if rect.thickness is not None:
        ax.add_patch(
            mpl_Rectangle(
                (
                    x - rect.half_width - rect.thickness,
                    y - rect.half_height - rect.thickness,
                ),
                2 * (rect.half_width + rect.thickness),
                2 * (rect.half_height + rect.thickness),
                **options_pml,
            )
        )



"""Polygon class."""


from __future__ import annotations

from dataclasses import dataclass, field
from fractions import Fraction
from math import gcd, lcm

from numpy import arctan2, asarray, greater, lexsort, pi
from numpy.linalg import norm
from numpy.typing import ArrayLike, NDArray

from .circular_iterable import circular_triplewise


class Angle(Fraction):
    """Angle class."""

    @property
    def value(self) -> float:
        """Return the angle value."""
        return float(self) * pi

    def __str__(self) -> str:
        if self.numerator == 0:
            return "0"

        if self.numerator == 1:
            str_num = ""
        elif self.numerator == -1:
            str_num = "-"
        else:
            str_num = f"{self.numerator}"

        return f"{str_num}π/{self.denominator}"

    __repr__ = __str__

    def latex(self) -> str:
        """Get latex representation."""
        if self.numerator == 0:
            return r"$0$"

        if self.numerator == 1:
            str_num = ""
        elif self.numerator == -1:
            str_num = "-"
        else:
            str_num = f"{self.numerator}"

        return rf"$\dfrac{{{str_num}\pi}}{{{self.denominator}}}$"


@dataclass(kw_only=True, slots=True)
class Corner:
    """Corner class."""

    c: NDArray
    angle: Angle
    v1: NDArray = field(repr=False)
    v2: NDArray = field(repr=False)

    def critical_interval(self) -> tuple[Fraction, Fraction]:
        """Get the critical interval."""
        a = (2 - self.angle) / self.angle
        b = 1 / a

        if self.angle > 1:
            return (-b, -a)

        return (-a, -b)

    def elementary_angle(self) -> Angle:
        """Compute the elementary angle of a corner.

        We compute p and q such that
        a / b = 2p / (p+q) and p, q >= 2 and return the angle 2/(p+q).

        """
        a, b = self.angle.numerator, self.angle.denominator
        if a % 2 == 0:
            p = a // 2
            q = b - p
        else:
            p = a
            q = 2 * b - a

        if (p == 1) or (q == 1):
            p, q = 2 * p, 2 * q

        return Angle(2, p + q)


@dataclass(kw_only=True, slots=True)
class Polygon:
    """Polygon class."""

    corners: list[Corner]
    name: str
    lengths: NDArray = field(repr=False)

    @classmethod
    def from_vertices(
        cls, vertices: ArrayLike, name: str, *, max_denominator: int = 32
    ) -> Polygon:
        """Create a polygon from the vertices."""
        pts = _fix_orientation(_to_array(vertices))

        lengths = _compute_lengths(pts)
        corners = _compute_corners(pts, max_denominator)

        return cls(corners=corners, name=name, lengths=lengths)

    @property
    def nbside(self) -> int:
        """Return the angle value."""
        return len(self.lengths)

    def translate(self, vector: NDArray) -> None:
        """Translate the vertices of the polygon."""
        for corner in self.corners:
            corner.c = corner.c + vector

    def get_vertices(self) -> NDArray:
        """Get verttices."""
        return asarray([corner.c for corner in self.corners])

    def get_elementary_angle(self) -> Angle:
        """Get the elementary angle."""
        angles = {c.elementary_angle() for c in self.corners}
        return Angle(
            gcd(*{a.numerator for a in angles}), lcm(*{a.denominator for a in angles})
        )

    def critical_interval(self) -> tuple[Fraction, Fraction]:
        """Get the critical interval."""
        a, b = self.corners[0].critical_interval()
        for corner in self.corners[1:]:
            I = corner.critical_interval()
            a = min(a, I[0])
            b = max(b, I[1])

        return (a, b)


def _to_array(vertices: ArrayLike) -> NDArray:
    """Check the shape of the vertices array."""
    pts = asarray(vertices)

    if len(pts.shape) != 2:
        raise ValueError("Vertices must be a two dimensional array.")

    nb_row, nb_col = pts.shape

    if not ((nb_row >= 3) and (nb_col == 2)):
        raise ValueError("vertices must have a shape (N, 2) with N ≥ 3.")

    return pts


def _fix_orientation(vertices: NDArray) -> NDArray:
    """Return the vertices in the counterclockwise direction."""
    n = vertices.shape[0]
    i = lexsort((vertices[:, 1], vertices[:, 0]))[0]

    C = vertices[(i - 1) % n, :]
    A = vertices[i, :]
    B = vertices[(i + 1) % n, :]
    AB = B - A
    AC = C - A

    if AB[0] * AC[1] - AB[1] * AC[0] < 0:
        return vertices[::-1, :]

    return vertices


def _compute_lengths(vertices: NDArray) -> NDArray:
    """Compute lengths."""
    index = [*range(vertices.shape[0]), 0]
    lengths = norm(vertices[index[1:], :] - vertices[index[:-1], :], axis=1)

    if not all(greater(lengths, 1e-8)):
        raise ValueError("There is two vertices too close.")

    return lengths


def _compute_corners(vertices: NDArray, max_denominator: int) -> list[Corner]:
    """Compute corners."""
    n = vertices.shape[0]

    corners = []
    for C, A, B in circular_triplewise(vertices, start=n - 1):
        AB = _normalize(B - A)
        AC = _normalize(C - A)
        angle = _compute_angle(AB, AC, max_denominator)
        corners.append(Corner(c=A, angle=angle, v1=AB, v2=AC))

    if sum(c.angle for c in corners) != n - 2:
        raise ValueError(
            "Error in the angle's computation at the polygon's corners, the issue "
            "might comes from:\n"
            "  • The angles are badly approximate by a rational times π.\n"
            "    - Possible solution is to increase max_denominator="
            f"{max_denominator}.\n"
            "  • The polygon is not simple."
        )

    return corners


def _normalize(vector: NDArray) -> NDArray:
    """Normalize vector."""
    return vector / norm(vector)


def _compute_angle(u: NDArray, v: NDArray, max_denominator: int) -> Angle:
    """Compute the angle between two vector."""

    _cos = u[0] * v[0] + u[1] * v[1]
    _sin = u[0] * v[1] - u[1] * v[0]
    angle = Fraction(arctan2(_sin, _cos) / pi).limit_denominator(max_denominator)

    if angle < 0:
        angle = 2 + angle

    return Angle(angle)



"""Geometry class."""


from __future__ import annotations

from dataclasses import dataclass
from fractions import Fraction
from typing import Iterable, Union

from numpy import asarray, vstack

from .border import AutoBorder, Border
from .closest_points import min_dist
from .polygon import Polygon


@dataclass(kw_only=True, slots=True)
class Geometry:
    """Geometry class.

    Attributes
    ----------
    polygons : list[Polygon]
    border : Border

    """

    polygons: list[Polygon]
    border: Border

    @classmethod
    def from_polygon(
        cls, polygon: Polygon, border: Union[Border, AutoBorder]
    ) -> Geometry:
        """Create a geometry from a polygon.

        Parameters
        ----------
        polygon : Polygon
        border : Union[Border, AutoBorder]

        """

        if isinstance(border, AutoBorder):
            return cls(
                polygons=[polygon], border=border.get_border(polygon.get_vertices())
            )

        if isinstance(border, Border):
            return cls(polygons=[polygon], border=border)

        raise ValueError("Unknown border shape.")

    @classmethod
    def from_polygons(
        cls, polygons: Iterable[Polygon], border: Union[Border, AutoBorder]
    ) -> Geometry:
        """Create a geometry from polygons.

        Parameters
        ----------
        polygon : Iterable[Polygon]
        border : Union[Border, AutoBorder]

        """

        points = vstack([polygon.get_vertices() for polygon in polygons])

        if isinstance(border, AutoBorder):
            return cls(polygons=list(polygons), border=border.get_border(points))

        if isinstance(border, Border):
            return cls(polygons=list(polygons), border=border)

        raise ValueError("Unknown border shape.")

    def make_center_origin(self):
        """Do a translation to set the center at (0, 0)."""
        center = self.border.center

        for polygon in self.polygons:
            polygon.translate(-center)

        self.border.center = asarray([0, 0])

    def critical_interval(self) -> dict[str, tuple[Fraction, Fraction]]:
        """Get the critical interval."""
        return {polygon.name: polygon.critical_interval() for polygon in self.polygons}

    def max_corner_radius(self) -> float:
        """Maximum corner radius."""
        points = vstack([polygon.get_vertices() for polygon in self.polygons])
        return min(min_dist(points) / 2, self.border.dist_to_inner_boundary(points))



"""Comptute the smallest enclosing circle or rectangle."""


from numpy import arange, asarray, lexsort, ones, sort, sqrt, stack
from numpy.linalg import det, norm
from numpy.random import shuffle
from numpy.typing import NDArray

from .circular_iterable import circular_triplewise

Circle = tuple[NDArray, float]

EPS_ADD: float = 1e-12
EPS_MUL: float = 1 + EPS_ADD


def convex_hull(points: NDArray) -> NDArray:
    """Computes the convex hull of a set of distinct 2D points.

    Implements Andrew's monotone chain algorithm from:
    A. M. Andrew, Another efficient algorithm for convex hulls in two dimensions
    https://doi.org/10.1016/0020-0190(79)90072-3
    and is heavly inspire by:
    https://en.wikibooks.org/wiki/Algorithm_Implementation/Geometry/Convex_hull/Monotone_chain

    Parameters
    ----------
    points : NDArray of shape (N, 2)
        set of 2D points

    Returns
    -------
    convex_hull : NDArray
        the points defining the convex hull

    """
    if (n := points.shape[0]) <= 1:
        return arange(n)

    idx_sort = lexsort((points[:, 1], points[:, 0]))

    lower = _lower_hull(points, idx_sort)
    upper = _lower_hull(points, idx_sort[::-1])

    return points[sort([*lower[:-1], *upper[:-1]]), :]


def _lower_hull(pts: NDArray, idx_sort: NDArray) -> list[int]:
    """Lower hull."""
    lower: list[int] = []
    for i in idx_sort:
        while (
            len(lower) >= 2
            and _cross_product(pts[lower[-2], :], pts[lower[-1], :], pts[i, :])
            <= EPS_ADD
        ):
            lower.pop()

        lower.append(i)

    return lower


def _cross_product(o: NDArray, u: NDArray, v: NDArray) -> float:
    """Cross product."""
    return (u[0] - o[0]) * (v[1] - o[1]) - (u[1] - o[1]) * (v[0] - o[0])


def smallest_circle(points: NDArray) -> Circle:
    """Compute the smallest enclosing circle of a collection of vertices using
    the Welzl's algorithm, see https://doi.org/10.1007/BFb0038202."""

    if points.shape[0] == 3:
        return _smallest_circle_3_points(*points)

    pts = convex_hull(points)
    shuffle(pts)

    return welzl(list(pts), [], pts.shape[0])


def welzl(points: list[NDArray], boundary: list[NDArray], lenght: int) -> Circle:
    """Welzl algorithm."""

    if lenght == 0 or len(boundary) == 3:
        return trivial_circle(boundary)

    P = points[lenght - 1]

    center, radius = welzl(points, boundary.copy(), lenght - 1)
    if _is_inside((center, radius), P):
        return (center, radius)

    boundary.append(P)
    return welzl(points, boundary.copy(), lenght - 1)


def trivial_circle(R: list[NDArray]) -> Circle:
    """Trivial circle."""
    match len(R):
        case 0:
            return (asarray([0, 0]), 0)

        case 1:
            return (R[0], 0)

        case 2:
            return _smallest_circle_2_points(R[0], R[1])

        case 3:
            return _smallest_circle_3_points(R[0], R[1], R[2])

        case _:
            raise ValueError("The length of R must be ≤ 3.")


def _smallest_circle_2_points(A: NDArray, B: NDArray) -> Circle:
    """Compute the smallest enclosing circle of two points."""

    center = asarray([(A[0] + B[0]) / 2, (A[1] + B[1]) / 2])
    radius = sqrt((A[0] - B[0]) ** 2 + (A[1] - B[1]) ** 2) / 2
    return (center, radius)


def _smallest_circle_3_points(A: NDArray, B: NDArray, C: NDArray) -> Circle:
    """Compute the smallest enclosing circle of three points."""

    for P, Q, R in circular_triplewise((A, B, C)):
        center, radius = _smallest_circle_2_points(P, Q)
        if _is_inside((center, radius), R):
            return (center, radius)

    return _circumcircle_triangle(A, B, C)


def _is_inside(circle: Circle, A: NDArray) -> bool:
    """Check if the point A is inside the circle."""
    C, r = circle
    return (A[0] - C[0]) ** 2 + (A[1] - C[1]) ** 2 < r**2 * EPS_MUL


def _circumcircle_triangle(A: NDArray, B: NDArray, C: NDArray) -> Circle:
    """Compute the circumcircle of a three points, see
    https://en.wikipedia.org/wiki/Circumscribed_circle."""
    points = asarray([A, B, C])

    norm2 = norm(points, axis=1) ** 2

    a = _det_col(points[:, 0], points[:, 1], ones(3))
    b = _det_col(points[:, 0], points[:, 1], norm2)

    cx = _det_col(norm2, points[:, 1], ones(3)) / (2 * a)
    cy = _det_col(points[:, 0], norm2, ones(3)) / (2 * a)

    return (asarray([cx, cy]), sqrt(b / a + cx * cx + cy * cy))


def _det_col(A: NDArray, B: NDArray, C: NDArray) -> float:
    """Compute the det of the matrix define by the colunm [A, B, C]."""
    return det(stack((A, B, C), axis=1))


def smallest_rectangle(points: NDArray) -> tuple[NDArray, NDArray]:
    """Find the smallest rectangle align with the coordinate axis."""
    xmin, xmax = points[:, 0].min(), points[:, 0].max()
    ymin, ymax = points[:, 1].min(), points[:, 1].max()

    center = asarray(((xmin + xmax) / 2, (ymin + ymax) / 2))
    lengths = asarray((xmax, ymax)) - center

    return (center, lengths)
